use axum::{
	extract::Query,
	extract::State,
	extract,
	response::Response,
	Router,
	routing::get,
};
use axum_client_ip::{SecureClientIp, SecureClientIpSource};

use tera::Tera;
use trust_dns_resolver::{
	TokioAsyncResolver,
	config::ResolverOpts,
	config::ResolverConfig,
};

use std::net::IpAddr;
use std::sync::Arc;
use std::path::Path;

mod simple_dns;
mod templating_engine;
mod geoip;
mod config;

use crate::geoip::QueryAsn;
use crate::geoip::QueryLocation;
use geoip::AsnResult;
use geoip::LocationResult;

use crate::templating_engine::View;
use crate::templating_engine::ResponseFormat;

#[derive(serde::Deserialize, serde::Serialize)]
pub struct BaseQuery {
	format: Option<ResponseFormat>,
	lang:   Option<String>,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct IpQuery {
	format: Option<ResponseFormat>,
	lang:   Option<String>,	
	ip:     IpAddr,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct DigQuery {
	format: Option<ResponseFormat>,
	name:   String,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct IpResult {
	hostname: Option<String>,
	asn:      Option<AsnResult>,
	location: Option<LocationResult>,
}


struct ServiceSharedState {
	templating_engine: templating_engine::Engine,
	dns_resolver:      TokioAsyncResolver,
	asn_db:            geoip::MMDBCarrier,
	location_db:       geoip::MMDBCarrier,
}


#[tokio::main]
async fn main() {
	// Initalize Tera templates
	// TODO: don't hardcode template directory
	println!("Parsing Templates ...");
	let res = Tera::new("templates/*.html");
	let tera = match res {
		Ok(t) => t,
		Err(e) => {
			println!("Template parsing error(s): {}", e);
			::std::process::exit(1);
		}
	};

	// Initalize GeoIP Database

	let mut asn_db = geoip::MMDBCarrier {
		mmdb: None,
		name: "GeoIP ASN Database".to_string(),
	};
	asn_db.load_database_from_path(Path::new("mmdb/GeoLite2-ASN.mmdb")).ok();

	let mut location_db = geoip::MMDBCarrier {
		mmdb: None,
		name: "GeoIP Location Database".to_string(),
	};
	location_db.load_database_from_path(Path::new("mmdb/GeoLite2-City.mmdb")).ok();
	
	// Initalize DNS resolver with os defaults
	println!("Initalizing dns resolver ...");
	let res = TokioAsyncResolver::tokio(ResolverConfig::default(), ResolverOpts::default());
	let dns_resolver = match res {
		Ok(resolver) => resolver,
		Err(e) => {
			println!("Error while setting up dns resolver: {e}");
			::std::process::exit(1);
		}
	};
	
	// Initialize shared state
	let shared_state = Arc::new(
		ServiceSharedState{
		templating_engine: templating_engine::Engine{
			tera: tera,
		},
		dns_resolver: dns_resolver,
		asn_db: asn_db,
		location_db: location_db,
	});

	// Initalize axum server
    let app = Router::new()
	    .route("/", get(handle_default_route))
	    .route("/dig", get(handle_dig_route))
	    .route("/dig/:name", get(handle_dig_route_with_path))
	    .route("/ip", get(handle_ip_route))
	    .route("/ip/:address", get(handle_ip_route_with_path))
    	.route("/hi", get(hello_world_handler))
    	.with_state(shared_state)
    	.layer(SecureClientIpSource::RightmostXForwardedFor.into_extension())
    	;

	println!("Starting Server ...");

    axum::Server::bind(&"0.0.0.0:3000".parse().unwrap())
        .serve(app.into_make_service())
        .await
        .unwrap();
}

async fn hello_world_handler(
	State(arc_state): State<Arc<ServiceSharedState>>,
) -> Response {
	let state = Arc::clone(&arc_state);

	state.templating_engine.render_view(
		ResponseFormat::TextPlain,
		View::Message("Hello! There, You, Awesome Creature!".to_string())
	).await	
}

async fn handle_default_route(
	Query(query): Query<BaseQuery>,
	State(arc_state): State<Arc<ServiceSharedState>>,
	SecureClientIp(address): SecureClientIp
) -> Response {

	let format = query.format.unwrap_or(ResponseFormat::TextHtml);

	let ip_query = IpQuery {
		format: query.format,
		lang: query.lang,
		ip: address,
	};
	
	let state = Arc::clone(&arc_state);

	// do reverse lookup
	let hostname = simple_dns::reverse_lookup(&state.dns_resolver, &address);

	// asn lookup
	let asn_result = state.asn_db.query_asn_for_ip(address);

	// location lookup
	let location_result = state.location_db.query_location_for_ip(
		address,
		&vec![&ip_query.lang.as_ref().unwrap_or(&"en".to_string()), &"en".to_string()]
	);

	let result = IpResult{
		hostname: hostname.await,
		asn: asn_result,
		location: location_result,
	};

	state.templating_engine.render_view(
		format,
		View::Index{query: ip_query, result: result}
	).await
}

async fn handle_ip_route(
	Query(ip_query): Query<IpQuery>,
	State(arc_state): State<Arc<ServiceSharedState>>,
) -> Response {
	return handle_ip_request(ip_query, arc_state).await
}

async fn handle_ip_route_with_path(
	Query(query): Query<BaseQuery>,
	State(arc_state): State<Arc<ServiceSharedState>>,
	extract::Path(address): extract::Path<IpAddr>,
) -> Response {	
	return handle_ip_request(IpQuery {
		format: query.format,
		lang: query.lang,
		ip: address,
	}, arc_state).await
}

async fn handle_ip_request(
	ip_query: IpQuery,
	arc_state: Arc<ServiceSharedState>,
) -> Response {

	let address = ip_query.ip;
	let format = ip_query.format.unwrap_or(ResponseFormat::TextHtml);

	let state = Arc::clone(&arc_state);

	// do reverse lookup
	let hostname = simple_dns::reverse_lookup(&state.dns_resolver, &address);

	// asn lookup
	let asn_result = state.asn_db.query_asn_for_ip(address);

	// location lookup
	let location_result = state.location_db.query_location_for_ip(
		address,
		&vec![&ip_query.lang.as_ref().unwrap_or(&"en".to_string()), &"en".to_string()]
	);

	let result = IpResult{
		hostname: hostname.await,
		asn: asn_result,
		location: location_result,
	};

	state.templating_engine.render_view(
		format,
		View::Ip{query: ip_query, result: result}
	).await
}

async fn handle_dig_route(
	Query(dig_query): Query<DigQuery>,
	State(arc_state): State<Arc<ServiceSharedState>>,
) -> Response {
	return handle_dig_request(dig_query, arc_state).await
}

async fn handle_dig_route_with_path(
	Query(query): Query<BaseQuery>,
	State(arc_state): State<Arc<ServiceSharedState>>,
	extract::Path(name): extract::Path<String>,
) -> Response {
	return handle_dig_request(DigQuery {
		format: query.format,
		name: name,
	}, arc_state).await
}

async fn handle_dig_request(
	dig_query: DigQuery,
	arc_state: Arc<ServiceSharedState>,
) -> Response {

	let state = Arc::clone(&arc_state);
	let name = &dig_query.name;
	let format = dig_query.format.unwrap_or(ResponseFormat::TextHtml);

	let dig_result = simple_dns::lookup(&state.dns_resolver, name, true).await;
	
	state.templating_engine.render_view(
		format,
		View::Dig{ query: dig_query, result: dig_result}
	).await

}
